'''
This utility is a database management system called AmazeDB.
It is a NoSQL, file-based DBMS.

The following classes are there:
    1. dbms.db(name):
        It defines a database of the name 'name'. More info 
        is available in the amazedb.dbms.db module.
    
    2. db.getGroup(name):
        It defines a data group of the name 'name'. More info
        is available in amazedb.dbms.group module.
'''
__all__ = ['dbms', 'group', 'run_tests']
